/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamfilter.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"


static gboolean    gdam_filter_is_ready_to_read (GdamSource*      source);
static int         gdam_filter_read             (GdamSource*	  filter,
                                       	         GdamSample*	  buf,
				       	         int		  n);
static gboolean    gdam_filter_seek             (GdamSource*      source,
                                                 GdamTime         offset,
		                                 int              whence);
static GdamTime    gdam_filter_get_position     (GdamSource*      source);

static void        gdam_filter_miss_data        (GdamSource*      input,
					         int              lost);

static gboolean    gdam_filter_detach_child     (GdamSource*      source,
		                                 GdamSource*      sub);
static void        gdam_filter_class_init       (GdamFilterClass* klass);
static void        gdam_filter_init             (GdamFilter*      filter);
static void        gdam_filter_destroy          (GdamFilter*      filter);
static void        gdam_filter_get_subsources   (GdamSource*      source,
                                                 int*             num_subs,
						 GdamSource***    subs);
static GtkObjectClass* parent_class;

GtkType gdam_filter_get_type()
{
	static GtkType filter_type = 0;
	if (!filter_type) {
		static const GtkTypeInfo filter_info = {
			"GdamFilter",
			sizeof(GdamFilter),
			sizeof(GdamFilterClass),
			(GtkClassInitFunc) gdam_filter_class_init,
			(GtkObjectInitFunc) gdam_filter_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
		filter_type = gtk_type_unique(GDAM_TYPE_SOURCE, &filter_info);
	}
	return filter_type;
}


static void gdam_filter_arg_set_filtered(GtkObject* object, GdamArg* arg)
{
	int source_id = GDAM_ARG_INT(arg);
	GdamSource* sub = gdam_namespace_local_source(source_id);
	gdam_debug("set_filtered: filter=%d, child=%d [source_id=%d]",
		GDAM_SOURCE(object)->source_id,
		sub ? GDAM_SOURCE(sub)->source_id : 0,
		source_id);
	gdam_filter_set_source(GDAM_FILTER(object), sub);
}
static void gdam_filter_arg_get_filtered(GtkObject* object, GdamArg* arg)
{
	GdamSource* filtered = GDAM_FILTER(object)->filtered;
	GDAM_ARG_INT(arg) = filtered ? filtered->source_id : 0;
}

static void gdam_filter_class_init(GdamFilterClass* klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	GdamSourceClass* source_class = (GdamSourceClass*)klass;
	klass->attach_filtered = NULL;
	source_class->read = gdam_filter_read;
	source_class->seek = gdam_filter_seek;
	source_class->get_position = gdam_filter_get_position;
	source_class->detach_child = gdam_filter_detach_child;
	source_class->miss_data = gdam_filter_miss_data;
	source_class->is_ready_to_read = gdam_filter_is_ready_to_read;
	source_class->get_subsources = gdam_filter_get_subsources;
	object_class->destroy = (void(*)(GtkObject*)) gdam_filter_destroy;
	gdam_arg_register(object_class,
	                  "filtered",
			  GDAM_ARG_TYPE_INT,
			  gdam_filter_arg_set_filtered,
			  gdam_filter_arg_get_filtered,
			  GDAM_ARG_VITAL);
}

static void gdam_filter_init(GdamFilter* filter)
{
	filter->filtered = NULL;
}

void gdam_filter_set_source(GdamFilter* filter, GdamSource* sub)
{
	GtkObjectClass* object_class = ((GtkObject*)filter)->klass;
	GdamFilterClass* filter_class = (GdamFilterClass*)object_class;
	g_return_if_fail((GdamSource*)filter != sub);

	{ int filt_id = GDAM_SOURCE(filter)->source_id;
	  int sub_id = sub ? sub->source_id : 0;
	  gdam_debug("gdam_filter_set_source: filter=%d, source=%d",
	  	filt_id, sub_id); }

		/* sub->ref_count ++ */
	if (sub)
		gtk_object_ref(GTK_OBJECT(sub));

	if (filter->filtered)
		gdam_source_detach_child(GDAM_SOURCE(filter),
		                         filter->filtered);
	if (!sub)
		return;

		/* sub->ref_count ++ */
	if (filter_class->attach_filtered)
		filter_class->attach_filtered(filter, sub);
	else
		gdam_source_orphan(sub);	/* increases sub's ref cnt */
	g_return_if_fail(!gdam_source_is_ancestor(GDAM_SOURCE(filter), sub));
	gtk_object_sink(GTK_OBJECT(sub));
	filter->filtered = sub;
	sub->parent = GDAM_SOURCE(filter);

		/* sub->ref_count -- */
	gtk_object_unref(GTK_OBJECT(sub));
}

GdamSource* gdam_filter_detach_source(GdamFilter* filter)
{
	GtkObjectClass* object_class = ((GtkObject*)filter)->klass;
	GdamSourceClass* source_class = (GdamSourceClass*)object_class;
	GdamSource* sub = filter->filtered;
	if (source_class->detach_child) {
		gtk_object_ref(GTK_OBJECT(sub));
		source_class->detach_child(GDAM_SOURCE(filter), sub);
	}
	filter->filtered = NULL;
	return sub;
}

void gdam_filter_kill_source(GdamFilter* filter)
{
	GtkObjectClass* object_class = ((GtkObject*)filter)->klass;
	GdamSourceClass* source_class = (GdamSourceClass*)object_class;
	GdamSource* sub = filter->filtered;
	//filter->filtered = NULL;
	source_class->detach_child(GDAM_SOURCE(filter), sub);
}

static gboolean gdam_filter_is_ready_to_read(GdamSource* source)
{
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered;
	g_return_val_if_fail(filter != NULL, FALSE);
	filtered = filter->filtered;
	return filtered ? gdam_source_is_ready_to_read(filtered)
	                : TRUE;
}
static int gdam_filter_read(GdamSource*	source, GdamSample* buf, int n)
{
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered;
	g_return_val_if_fail(filter != NULL, 0);
	filtered = filter->filtered;
	if (filtered == NULL)
		return 0;
	return gdam_source_read(filtered, buf, n);
}
static gboolean gdam_filter_seek(GdamSource*      source,
                                 GdamTime         offset,
		                 int              whence)
{
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered;
	g_return_val_if_fail(filter != NULL, FALSE);
	filtered = filter->filtered;
	if (filtered == NULL) {
		g_warning("filter-seek on filter without subsource");
		return FALSE;
	}
	return gdam_source_seek(filtered, offset, whence);
}

static GdamTime gdam_filter_get_position(GdamSource*      source)
{
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered;
	g_return_val_if_fail(filter != NULL, FALSE);
	filtered = filter->filtered;
	if (filtered == NULL) {
		g_warning("filter-getposition on filter without subsource");
		return FALSE;
	}

	/* Check for circular dependency. */
	g_return_val_if_fail(filtered != source, -1);
	return gdam_source_get_position(filtered);
}
static void gdam_filter_miss_data(GdamSource* source, int lost)
{
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered;
	g_return_if_fail(filter != NULL);
	filtered = filter->filtered;
	if (!filtered)
		return;
	gdam_source_miss_data(filtered, lost);
}

static gboolean gdam_filter_detach_child(GdamSource* source, GdamSource* sub)
{
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered = filter->filtered;
	g_return_val_if_fail(filter != NULL, FALSE);

	if (filter->filtered == NULL)
		return FALSE;

	g_return_val_if_fail(filter->filtered == sub, FALSE);

	filter->filtered = NULL;
	filtered->parent = NULL;
	gtk_object_unref(GTK_OBJECT(filtered));

	return TRUE;
}
static void        gdam_filter_get_subsources   (GdamSource*      source,
                                                 int*             num_subs,
						 GdamSource***    subs)
{
	GdamFilter* filter = GDAM_FILTER(source);
	g_return_if_fail(filter != NULL);
	if (filter->filtered) {
		*num_subs = 1;
		*subs = g_new(GdamSource*, 1);
		(*subs)[0] = filter->filtered;
	} else {
		*num_subs = 0;
		*subs = NULL;
	}
}

static void gdam_filter_destroy(GdamFilter* filter)
{
	if (filter->filtered)
		gdam_filter_kill_source(filter);
	parent_class->destroy((GtkObject*)filter);
}

int         gdam_filter_sub_read       (GdamFilter* filter, 
                                        GdamSample* buf,
					int         n)
{
	if (filter->filtered == NULL)
		return 0;
	return gdam_source_read(filter->filtered, buf, n);
}
