/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamwalker.h"
#include <gdam/gdam-protocol.h>
#include <stdlib.h>
#include <math.h>

static int         gdam_walker_read             (GdamSource*   input,
	                                         GdamSample*   samples,
						 int           max_read);
static void        gdam_walker_miss_data        (GdamSource*   input,
	                                         int           samples_lost);
static void        gdam_walker_destroy          (GtkObject*    object);
static gboolean    gdam_walker_setup            (GdamWalker*   walker,
                                                 guint         buffer_id);
static GtkObjectClass* parent_class = NULL;
static GdamSourceClass* parent_source_class = NULL;

static inline double gdam_walker_drand48(GdamWalker *this)
{
	return erand48(this->random_state);
}
static inline void gdam_walker_srand48(GdamWalker *this, int seed)
{
	this->random_state[0] = 0x330e;
  	*((int *) &this->random_state[1]) = seed;
}
static inline int gdam_walker_get_seed(GdamWalker *this)
{
  	return *((int *) &this->random_state[1]);
}

static void gdam_walker_set_mean_length(GtkObject* object, GdamArg* arg)
{
	GDAM_WALKER(object)->mean_length = GDAM_ARG_DOUBLE(arg);
}
static void gdam_walker_get_mean_length(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_DOUBLE(arg) = GDAM_WALKER(object)->mean_length;
}
static void gdam_walker_set_random_seed(GtkObject* object, GdamArg* arg)
{
  	gdam_walker_srand48(GDAM_WALKER(object), GDAM_ARG_INT(arg));
}
static void gdam_walker_get_random_seed(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = gdam_walker_get_seed(GDAM_WALKER(object));
}
static void gdam_walker_set_buffer_id(GtkObject* object, GdamArg* arg)
{
	GdamWalker* walker = GDAM_WALKER(object);
	if (!gdam_walker_setup(walker, GDAM_ARG_INT(arg))) {
		gdam_debug("walker_setup failed (buffer_id=%d)",
			GDAM_ARG_INT(arg));
		return;
	}
	if (walker->position >= walker->cache->buffer_length)
		walker->position = walker->cache->buffer_length - 1;
}
static void gdam_walker_get_buffer_id(GtkObject* object, GdamArg* arg)
{
	GdamCacheBuffer* buffer = GDAM_WALKER(object)->cache;
	GDAM_ARG_INT(arg) = buffer ? buffer->buffer_id : 0;
}


static void gdam_walker_class_init(GdamSourceClass* to_init)
{
	GtkObjectClass* object_class = (GtkObjectClass*)to_init;
	to_init->read = gdam_walker_read;
	to_init->miss_data = gdam_walker_miss_data;
	object_class->destroy = gdam_walker_destroy;

	gdam_arg_register(object_class,
	                  "mean_length",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_walker_set_mean_length,
			  gdam_walker_get_mean_length,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "random_seed",
			  GDAM_ARG_TYPE_INT,
			  gdam_walker_set_random_seed,
			  gdam_walker_get_random_seed,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "buffer_id",
			  GDAM_ARG_TYPE_INT,
			  gdam_walker_set_buffer_id,
			  gdam_walker_get_buffer_id,
			  GDAM_ARG_VITAL);
}

static void gdam_walker_init(GdamWalker* to_init)
{
	to_init->mean_length = 441;
	gdam_walker_srand48(to_init, 0xdeadbeef);
}

GtkType gdam_walker_get_type ()
{
	static GtkType walker_type = 0;
	if (!walker_type) {
		static const GtkTypeInfo walker_info = {
			"GdamWalker",
			sizeof(GdamWalker),
			sizeof(GdamSourceClass),
			(GtkClassInitFunc) gdam_walker_class_init,
			(GtkObjectInitFunc) gdam_walker_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_SOURCE);
		parent_source_class = (GdamSourceClass*)parent_class;
		walker_type = gtk_type_unique(GDAM_TYPE_SOURCE,
		                                  &walker_info);
	}
	return walker_type;
}

static gboolean    gdam_walker_setup(GdamWalker*    walker,
                                        guint             buffer_id)
{
	GdamCacheBuffer* cache_buf;
	cache_buf = gdam_cache_buffer_get_by_id(buffer_id);
	if (cache_buf->buffer_length <= 0) {
		gdam_debug("walker %u: still loading...", buffer_id);
		return FALSE;
	}
	gtk_object_ref(GTK_OBJECT(cache_buf));
	if (walker->cache)
		gtk_object_unref(GTK_OBJECT(walker->cache));
	walker->cache = cache_buf;
	return TRUE;
}

GdamWalker* gdam_walker_new(guint buffer_id)
{
	GdamWalker* walker;
	walker = GDAM_WALKER(gtk_type_new(GDAM_TYPE_WALKER));
	if (!gdam_walker_setup(walker, buffer_id)) {
		gtk_object_destroy(GTK_OBJECT(walker));
		return NULL;
	}
	return walker;
}

static int       gdam_walker_read           (GdamSource*      input,
	                                        GdamSample*      samples,
						int              max_read)
{
	GdamWalker* walker = GDAM_WALKER(input);
	int buffer_len = walker->cache->buffer_length;
	GdamSample* buffer = walker->cache->buffer;
	int to_read = max_read;
	int position = walker->position;
	int direction = walker->direction;
	int length = walker->length;

	//gdam_debug("walker_read: %d bytes [start offset=%d]", max_read, position);

	while (to_read > 0) {
		double x;
		while (to_read > 0 && length > 0) {
			if (direction == -1) {
				if (--position < 0) {
					position = 0;
					length++;
					direction = +1;
				}
			} else {
				if (++position >= buffer_len) {
					position = buffer_len-1;
					length++;
					direction = -1;
				}
			}
			*samples++ = buffer[position];
			to_read--;
			length--;
		}
		if (length == 0) {
			do
				x = gdam_walker_drand48(walker);
			while (x == 0.0);
			length = (int) (-walker->mean_length * log(x));
			direction =
				(gdam_walker_drand48(walker) < 0.5 ? -1 : 1);
		}
	}
	walker->position = position;
	walker->direction = direction;
	walker->length = length;

	return max_read;
}

static void      gdam_walker_miss_data      (GdamSource*      input,
	                                        int              samples_lost)
{
	//GdamWalker* walker = GDAM_WALKER(input);
}

static void      gdam_walker_destroy            (GtkObject*       object)
{
	GdamWalker* walker = GDAM_WALKER(object);
	if (walker->cache) {
		GtkObject* buf = GTK_OBJECT(walker->cache);
		walker->cache = NULL;
		gtk_object_unref(buf);
	}
	parent_class->destroy(object);
}
